from typing import Mapping, Any

from toya.supplier.supplier import Supplier


class MappingSupplier(Supplier):

    def __init__(self, values: Mapping[str, Any]):
        self.values = values

    def get(self) -> Mapping[str, Any]:
        return self.values



import functools

import yaml
from yaml import YAMLObject
from yaml.nodes import ScalarNode

DEFAULT_TAG = "!t"


class Tpl(YAMLObject):
    value: str

    yaml_flow_style = True
    yaml_loader = [yaml.SafeLoader, yaml.Loader, yaml.FullLoader, yaml.UnsafeLoader]

    value: str

    def __init__(self, value: str):
        self.value = value

    def __repr__(self) -> str:
        return f"Tpl({self.value})"

    @classmethod
    def from_yaml(cls, loader, node):
        if not isinstance(node, ScalarNode):
            raise TypeError(f"Expected ScalarNode, got {type(node)}")
        if not isinstance(node.value, str):
            raise TypeError(f"Expected str, got {type(node.value)}")
        return cls(node.value)

    @classmethod
    def to_yaml(cls, dumper, data):
        return dumper.represent_scalar(cls.yaml_tag, data.value)


@functools.cache
def _tpl_type_constructor(tag: str) -> type[Tpl]:
    class TplImpl(Tpl):
        yaml_tag = tag

    return TplImpl


def create_tpl_type(tag: str = DEFAULT_TAG) -> type[Tpl]:
    return _tpl_type_constructor(tag)



from collections import OrderedDict
from typing import Mapping, Any, List, MutableMapping

from jinja2 import Environment
from mergedeep import merge

from toya.supplier.supplier import Supplier
from toya.tpl import Tpl


class ConfigError(Exception):
    pass


def load_config(suppliers: List[Supplier]) -> MutableMapping[str, Any]:
    res: MutableMapping[str, Any] = OrderedDict()
    for supplier in suppliers:
        data = supplier.get()
        if not isinstance(data, Mapping):
            raise ConfigError(f"data is not a mapping; {type(data)=}")
        merge(res, data)
    return res


def eval_config(config: MutableMapping[str, Any]) -> None:
    env = Environment()

    def eval_cfg(cfg: MutableMapping[Any, Any]) -> None:
        for k, v in cfg.items():
            if isinstance(v, Tpl):
                tpl = env.from_string(v.value, {"_": config})
                rendered_value = tpl.render(cfg)
                cfg[k] = rendered_value
            elif isinstance(v, dict):
                eval_cfg(v)

    eval_cfg(config)


def load_and_eval_config(suppliers: List[Supplier]) -> MutableMapping[str, Any]:
    config = load_config(suppliers)
    eval_config(config)
    return config



import os
from typing import Generator

import pytest


@pytest.fixture
def preserve_env() -> Generator[None]:
    store = dict()
    for k, v in os.environ.items():
        store[k] = v
    yield
    for k, v in os.environ.items():
        if k not in store:
            del os.environ[k]
        elif os.environ[k] != v:
            os.environ[k] = v



import os

from common import resource_path
from toya.config import load_and_eval_config
from toya.supplier.env_supplier import EnvSupplier
from toya.supplier.mapping_supplier import MappingSupplier
from toya.supplier.yaml_supplier import YamlSupplier
from toya.tpl import create_tpl_type


def test_check_type_cache():
    assert create_tpl_type("!test1") != create_tpl_type("!test2")
    assert create_tpl_type("!test") == create_tpl_type("!test")


def test_yaml_with_default_tag():
    cfg = load_and_eval_config([YamlSupplier(resource_path("test1.yaml"))])
    assert cfg["c1"] == "v1"
    assert cfg["c2"] == "v1/v2"


def test_yaml_with_custom_tag():
    cfg = load_and_eval_config([YamlSupplier(resource_path("test2.yaml"), "!tpl")])
    assert cfg["c1"] == "v1"
    assert cfg["c2"] == "v1/v2"


def test_env(preserve_env):
    os.environ["APP__C1"] = "e1"
    cfg = load_and_eval_config([
        YamlSupplier(resource_path("test3.yaml")),
        EnvSupplier()
    ])
    assert cfg["c1"] == "e1"
    assert cfg["c2"] == "e1/v2"


def test_yaml_props_ref_and_env_and_mapping(preserve_env):
    os.environ["APP__C2"] = "e2"
    os.environ["APP__C3__C32"] = "e32"
    tpl_type = create_tpl_type()
    cfg = load_and_eval_config([
        YamlSupplier(resource_path("test4.yaml")),
        EnvSupplier(),
        MappingSupplier({
            "c3": {
                "c32": tpl_type("{{ _.c1 }}@{{ c31 }}")
            },
        })
    ])
    assert cfg["c3"]["c33"] == "v1/e2/v31/v1@v31"



from pathlib import Path


def resource_path(relative_path: str | Path | None = None) -> Path:
    res = (Path(__file__).parent / "resources").resolve()
    if relative_path is None:
        return res
    return res / relative_path



from pathlib import Path
from typing import Mapping, Any

import yaml

from toya.supplier.supplier import Supplier
from toya.tpl import create_tpl_type, DEFAULT_TAG


class YamlSupplier(Supplier):

    def __init__(self, path: Path, tag: str = DEFAULT_TAG) -> None:
        self.path = path
        self.tag = tag

    def get(self) -> Mapping[str, Any]:
        create_tpl_type(self.tag)
        with self.path.open() as stream:
            document: Mapping[str, Any] = yaml.safe_load(stream)
        return document



import os
from typing import Mapping, Any

from toya.supplier.supplier import Supplier


class EnvSupplier(Supplier):

    def __init__(self, prefix: str = "APP__", separator: str = "__"):
        self.prefix = prefix
        self.separator = separator

    def get(self) -> Mapping[str, Any]:
        def update(d: dict[str, Any], path: list[str], ind: int, v: Any) -> None:
            last = len(path) - 1
            if last < 0:
                return
            key = path[ind]
            if ind == last:
                d[key] = v
            else:
                if key not in d:
                    d[key] = {}
                update(d[key], path, ind + 1, v)

        res: dict[str, Any] = {}
        for key, value in os.environ.items():
            if key and key.startswith(self.prefix):
                path = [k.lower() for k in key[len(self.prefix):].split(self.separator)]
                update(res, path, 0, value)
        return res
